'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getInlinedOperatorsEnabled = getInlinedOperatorsEnabled;
exports.replaceWithInlinedOperator = replaceWithInlinedOperator;

var _types = require('./types');

var _types2 = _interopRequireDefault(_types);

var _astLocUtils = require('ast-loc-utils');

var _compilerState = require('./compilerState');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var inlinedOperator = {
  looseEq: function looseEq(args) {
    return _types2.default.binaryExpression("==", args[0], args[1]);
  },
  looseNotEq: function looseNotEq(args) {
    return _types2.default.binaryExpression("!=", args[0], args[1]);
  },
  bitwiseNot: function bitwiseNot(args) {
    return _types2.default.unaryExpression("~", args[0]);
  },
  bitwiseAnd: function bitwiseAnd(args) {
    return _types2.default.binaryExpression("&", args[0], args[1]);
  },
  bitwiseOr: function bitwiseOr(args) {
    return _types2.default.binaryExpression("|", args[0], args[1]);
  },
  bitwiseXor: function bitwiseXor(args) {
    return _types2.default.binaryExpression("^", args[0], args[1]);
  },
  bitwiseLeftShift: function bitwiseLeftShift(args) {
    return _types2.default.binaryExpression("<<", args[0], args[1]);
  },
  bitwiseRightShift: function bitwiseRightShift(args) {
    return _types2.default.binaryExpression(">>", args[0], args[1]);
  },
  bitwiseZeroFillRightShift: function bitwiseZeroFillRightShift(args) {
    return _types2.default.binaryExpression(">>>", args[0], args[1]);
  }
};

function getInlinedOperatorsEnabled(opts) {
  return !(opts.stdlib === false || typeof opts.stdlib === "object" && opts.stdlib.lightscript === false);
}function replaceWithInlinedOperator(path, callee, args) {
  if (!(0, _compilerState.getCompilerState)().inlinedOperatorsEnabled || callee.type !== "Identifier") return false;
  var name = callee.name;
  if (!inlinedOperator[name] || path.scope.hasBinding(name)) return false;
  path.replaceWith((0, _astLocUtils.placeAtNode)(inlinedOperator[name](args), path.node));
  return true;
}