"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.transformPlaceholders = transformPlaceholders;

var _types = require("./types");

var _types2 = _interopRequireDefault(_types);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function registerPlaceholder(functions, path, programPath) {
  var _ref;

  var fnPath = path.getFunctionParent();
  if (fnPath === programPath || !fnPath.node || !fnPath.node.params) {
    throw path.buildCodeFrameError("Placeholders cannot be used outside functions.");
  }if (fnPath.node.params.length) {
    throw path.buildCodeFrameError("Placeholders cannot be used in functions with arguments.");
  }if ((((_ref = path.parentPath) == null ? null : _ref.node) == null ? null : _ref.node.type) === "SpreadElement") {
    var _ref2;

    if (((_ref2 = path.node) == null ? null : _ref2.index) != null) {
      throw path.buildCodeFrameError("Cannot use indices with spread placeholders.");
    }
  }var fnInfo = functions.get(fnPath);
  if (fnInfo) {
    return fnInfo.placeholders.push(path);
  } else {
    return functions.set(fnPath, { functionPath: fnPath, placeholders: [path] });
  }
}function liftPlaceholders(fnPath, phPaths) {
  // Collate placeholders
  var ixPhPaths = phPaths.filter(function (x) {
    return x.node.index != null;
  });
  var nixPhPaths = phPaths.filter(function (x) {
    return !(x.node.index != null);
  });
  ixPhPaths.sort(function (a, b) {
    return a.node.index - b.node.index;
  });
  nixPhPaths.sort(function (a, b) {
    var _ref3, _ref4;

    // SpreadElements are last
    if (((_ref3 = a.parentPath) == null ? null : _ref3.node.type) === "SpreadElement") {
      return 1;
    } else if (((_ref4 = b.parentPath) == null ? null : _ref4.node.type) === "SpreadElement") {
      return -1;
    } else {
      return 0;
    }
  });
  var sortedPhPaths = ixPhPaths.concat(nixPhPaths);

  // Formulate args
  var args = [];
  var prevIndex = undefined,
      restArg = undefined;
  for (var _i2 = 0, _len2 = sortedPhPaths.length; _i2 < _len2; _i2++) {
    var _ref5;

    var _phPath = sortedPhPaths[_i2];var _node = _phPath.node;

    if (((_ref5 = _phPath.parentPath) == null ? null : _ref5.node.type) === "SpreadElement") {
      if (!restArg) {
        restArg = fnPath.scope.generateUidIdentifier("arg");
        args.push(_types2.default.restElement(restArg));
      }_phPath.replaceWith(restArg);
    } else if (_node.index != null && _node.index === prevIndex) {
      // Repeated use of an index = same arg.
      _phPath.replaceWith(args[args.length - 1]);
    } else {
      args.push(fnPath.scope.generateUidIdentifier("arg"));
      prevIndex = _node.index;
      _phPath.replaceWith(args[args.length - 1]);
    }
  } // Lift args to function level
  var fnNode = fnPath.node;
  fnNode.params = args;
  return fnPath.replaceWith(fnNode);
} // Collect all functions with placeholders into a Map, one entry per function
function findFunctionsWithPlaceholders(programPath) {
  var fns = new Map();
  programPath.traverse({
    PlaceholderExpression(path) {
      registerPlaceholder(fns, path, programPath);
    } });
  return fns;
}function transformPlaceholders(programPath) {
  var fns = findFunctionsWithPlaceholders(programPath);
  return fns.forEach(function (fnInfo) {
    return liftPlaceholders(fnInfo.functionPath, fnInfo.placeholders);
  });
}